//============================================================================================================================
//HermIRES - Hires Bitmap Editor for C64 - Hermit Software Hungary (Mihaly Horvath)     -----      Year 2012 A.D.
//============================================================================================================================
#include <allegro.h>
#include <math.h>
#include <map>
#include <string.h>
#include <ctype.h>

//using namespace std;
//--------------------------------------global variable declarations-----------------------------------------------------

/* XPM program-icon*/
static const char *allegico_xpm[] = {
/* columns rows colors chars-per-pixel */
"32 32 8 1 ",
"  c black",
". c #494949494949",
"X c #6D6D6D6D6D6D",
"o c #B6B600000000",
"O c #B6B624242424",
"+ c #6D6D6D6DDBDB",
"@ c #6D6D6D6DFFFF",
"# c #B6B6B6B6B6B6",
/* pixels */
"                                ",
"                                ",
"   .       .       .       .    ",
"   X       X@@@@@@@X       X    ",
"   #     @@#@@@@@@@#       #    ",
"   #    @@@#@@@@@@@#       #    ",
" .###########################.  ",
"   #  @@@@@#@@@@@@@#       #    ",
"   # @@@@@@#@@@@@@@#       #    ",
"   # @+@@@@#       #@@++++@#    ",
"   #++++++ #       #+++++++#    ",
"   #@@@@@@ #       #@@@@@@ #    ",
"   #@@@@@@ #       #@@@@@  #    ",
"   #@@@@@@ #       #@@@@   #    ",
" .###########################.  ",
"   #@@@@@@ #       #oooo   #    ",
"   #@@@@@@ #       #ooooo  #    ",
"   #@@@@@@ #       #oooooo #    ",
"   #++++++ #       #OOOOOOO#    ",
"   # @@@@@@#       #ooooooo#    ",
"   # @@@@@@#@@@@@@@#       #    ",
"   #  @@@@@#@@@@@@@#       #    ",
" .###########################.  ",
"   #    @@@#@@@@@@@#       #    ",
"   #     @@#@@@@@@@#       #    ",
"   #      @#@@@@@@@#       #    ",
"   #       #       #       #    ",
"   #       #       #       #    ",
"   X       X       X       X    ",
"   .       .       .       .    ",
"                                ",
"                                "
};
#if defined ALLEGRO_WITH_XWINDOWS && defined ALLEGRO_USE_CONSTRUCTOR
extern void *allegro_icon;
CONSTRUCTOR_FUNCTION(static void _set_allegro_icon(void));
static void _set_allegro_icon(void)
{
    allegro_icon = allegico_xpm;
}
#endif


BITMAP *DrawGrid, *Preview, *ToolBar;
PALETTE BMPalette; //no need for save, we use true-colour - for opening&converting 256 colour bitmaps, this may be needed
unsigned int C64Palette[16] = {0x000000,0xFFFFFF,0x68372B,0x70A4B2,0x6F3D86,0x588D43,0x352879,0xB8C76F,
                               0x6F4F25,0x433900,0x9A6759,0x444444,0x6C6C6C,0x9AD284,0x6C5EB5,0x959595};
std::map <unsigned int, unsigned int> BackPalette; //will be used for back-conversion from bmp....
const char* Hex[16]={"0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"};
const unsigned char pow2[8] = {128,64,32,16,8,4,2,1}; //powers of 2 precalculated for fast conversion
int ChrSizeX=8, ChrSizeY=8, NoOfChrX=40, NoOfChrY=25;
int FieldSizeX = ChrSizeX*NoOfChrX; //SprSizeX*NoOfSprX; 
int FieldSizeY = ChrSizeY*NoOfChrY; //SprSizeY*NoOfSprY;
int Mag_min=3, Mag_max=32, MagnifyFactor = Mag_min;
int GridBorder=10, GridPosX = GridBorder, GridPosY = GridBorder, GridSizeX = Mag_min*FieldSizeX, GridSizeY = Mag_min*FieldSizeY, GridCol = 0x202020,ChrRangeCol = 0xC0C0C0;
int WheelBuf = MagnifyFactor, MButtBuf=0, PreviewPosX=GridBorder, PreviewPosY=GridPosY+GridSizeY+GridBorder; 
int GuiSizeX=FieldSizeX*MagnifyFactor+GridBorder*2, GuiSizeY=FieldSizeY+PreviewPosY+GridBorder; 
int PalButtPosX=FieldSizeX+GridBorder*2, PalButtPosY=GridSizeY+GridBorder*2+10, PalButtSizeX = (GuiSizeX-PalButtPosX-GridBorder)/16, PalButtSizeY=GridSizeY/16;
unsigned char BmpData [328][208]; //dimensions (FieldSizeX, FieldSizeY)
unsigned char BmpClipBoard [320][200]; //dimensions (FieldSizeX, FieldSizeY)
unsigned char C64bmp[0x2000];
int ClipBoardSizeX=0, ClipBoardSizeY=0;
unsigned char ConvChr[16][9]; //storing result here for 1 calculated Char on conversion (function ConvertChr)
unsigned char ClashCnt[1040]; //colour-count of characters to determine clash
unsigned char ColorMap[1024]; //colour-map for the hires bitmap
unsigned char CopyBuffer[8][8]; //buffer for copy
unsigned char BorderColWidth=4, DefaultBorderCol=11, DefaultBGcol=12, BorderCol=DefaultBorderCol, DrawCol1=1, DrawCol2=DefaultBGcol, DitherCol=14, SelCol=0; 
int CurPosX = 10, CurPosY = 10, GridInPosX=0, GridInPosY=0, ActChar=0, ActQuad=0, ActChrBuf=0; //default position & colours
int mickeyx,mickeyy,TempPosX,TempPosY,TempPosX2,TempPosY2, TempStartX,TempStartY, PrevRightClick=0;
const int ControlTimer = 20;
volatile int CloseButtState=FALSE; int RButt=0;
char path[480]={"untitled.hbm"};
unsigned int PrgStartAdd = 0x2000;  //Char-data loadaddress in hexadecimal
unsigned int BASICstartAdd = 0x0801; //EXE export's BASIC starter load-address
unsigned int EXEdataAdd = 0x0A00; //EXE export's data-position
unsigned char PreviewZoom = 1; //zooming ratio of preview
int PenDiameter=4, TBSizeY=22, TButtSize=TBSizeY, NoOfButt = 8, TBSizeX=TButtSize*NoOfButt+1, 
    TBCol= 0xA0A0A0, TBFGCol = 0x20202020, TBSelCol = 0xffe080;
int TBPosX = PreviewPosX+FieldSizeX+30,TBPosY = PreviewPosY+(GuiSizeY-PreviewPosY)/2-10;
int TxtInfoPosX=TBPosX+TBSizeX+30, TxtInfoPosY=PalButtPosY+PalButtSizeY+30;
int MenuPosX=GuiSizeX-200, MenuPosY=PalButtPosY+PalButtSizeY+30;
int LinerSwitch = 1, DrawMode=0, RectFill=0, CircFill=0, DitherMode=0, Dither2Col=0; 
float pi=3.1415926, rad=pi/180, GridBrightness=1.00;
unsigned char UndoStore[16][321][201]; //16 full hbm pictures can be stored for undo/redo in this ring-buffer array
unsigned char UndoPointer = 0, UndoBase = 0, RedoMax=0; //undobase counts from 0..15 whenever UndoPointer reaches it, loosing oldest step at that time
int ShapeDrawn = 0, mousestartx, mousestarty;
int ShapeRightButt=0;
int ClashTimer=0, ClashBlinkSpeed=48, ClashBlinkRatio=8, BlinkOn=0;
bool ClashTestMode=true, EnableGrid=true, EnableBriGrid=true, WideMode=false;
int mouse_x_buf, mouse_y_buf;

//function prototypes
void InitBmp(); //InitSpr(); 
void SetTimer();
void GetControls();
void BmpDrawer();
void GridDrawer(int GridCol);
void PrevDrawer();void DrawPrv();
void DrawGridPixel(int x, int y, int col, int chkChar);
void DrawPrevPixel(int x, int y, int col);
void SetBackRange(); int CheckRange(); int GetDataX(); int GetDataY();
void CloseButton(); int Quit(void); int Help(void);
int my_button_proc(int msg, DIALOG *d, int c);
int Load();int Save();int ChangeCol();
int SelCol0(); int SelCol1(); int SelCol2(); int SelCol3(); int SelCol4(); int SelCol5(); int SelCol6(); int SelCol7(); int SelCol8(); int SelCol9();
int SelColA(); int SelColB(); int SelColC(); int SelColD(); int SelColE(); int SelColF();
void SelectChar(int num); int Pluser(); int Minuser(); void Zoomer(); void TextDispl(); void CalcActChar();
int ClearBmp(); void DispSelCol(); int Spacer(); int ExportC64(); int ExportExe(); int Deleter(); int CopyChar(); int Paster(); int Cutter(); 
int Homer(); int Ender(); int PageUp(); int PageDn(); int Lefter(); int Righter(); int Upper(); int Downer(); void PrevText();
int SelChrPosX(int Chrnum); 
int SelChrPosY(int Chrnum);
int SaveBMP(); void LoadFile(); void LoadBMP();
void DrawBmpData(int x, int y, unsigned char colour);
unsigned int GetMatchCol(unsigned int inputcol);
int F1er(); int F2er(); int F3er(); int F4er(); int F5er();int F6er();int F7er();int F8er();int F9er();int F10er();int F11er();int Zeer(); int Yeer();
void DispToolBar(); void BuffPos();
void DrawChrPixel(int x, int y, unsigned char col);
void DrawBmpLine(int x1, int y1, int x2, int y2, unsigned char col);
void Undoer(); void Redoer(); void StoreUndo();
int ConvertChr(int Charnum);
void ConvBackBMP(BITMAP *bitmap);
void LineShape(); //draw a line from muse-click position to mouse-release position
void RectDraw(); //draw a rectangle (normal/filled) from muse-click position to mouse-release position
void CircleDraw();
void Filler();
void CopyPasteSelection();
int ClashTest(), ToggleGrid(), ToggleBriGrid(), ToggleWide();
void DetectClashes();
void BorderDrawer();
void DisplayClipBoard();

DIALOG OverDialog [] =  //Array of GUI-dialog elements
{
 /* (proc)        (x)  (y)  (w) (h) (fg) (bg) (key) (flags) (d1) (d2) (dp)           (dp2) (dp3) */
//Clear Dialog screen
 { d_clear_proc,        0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  },
//Main Buttons
 { my_button_proc, MenuPosX+100, MenuPosY+100, 85, 20, 0xF0F0F0, 0x808080,  'q', D_EXIT, 16, 0, (void*)"&Quit", NULL, (void *)Quit},
 { my_button_proc, MenuPosX+100, MenuPosY    , 85, 20, 0xF0F0F0, 0x808080,  's', D_EXIT, 16, 0, (void*)"&Save", NULL, (void *)Save},
 { my_button_proc, MenuPosX+100, MenuPosY+35 , 85, 20, 0xF0F0F0, 0x808080,    0, D_EXIT, 16, 0, (void*)"Export C64", NULL, (void*)ExportC64},
 { my_button_proc, MenuPosX+100, MenuPosY+65 , 85, 20, 0xF0F0F0, 0x808080,    0, D_EXIT, 16, 0, (void*)"Export Exe", NULL, (void*)ExportExe},
 { my_button_proc, MenuPosX    , MenuPosY    , 85, 20, 0xF0F0F0, 0x808080, 'l', D_EXIT, 16, 0, (void*)"&Load", NULL, (void *)Load},
 { my_button_proc, MenuPosX    , MenuPosY+50 , 85, 20, 0xF0F0F0, 0x808080,   0, D_EXIT, 16, 0, (void*)"Clear", NULL, (void *)ClearBmp},
 { my_button_proc, MenuPosX    , MenuPosY+100, 85, 20, 0xF0F0F0, 0x808080, 'h', D_EXIT, 16, 0, (void*)"&Help", NULL, (void *)Help},

 //{ my_button_proc, MenuPosX-400, MenuPosY+100, 85, 20, 0xF0F0F0, 0x808080, 'w', D_EXIT, 16, 0, (void*)"&Width", NULL, (void *)ToggleWide},
 { my_button_proc, MenuPosX-300, MenuPosY+100, 85, 20, 0xF0F0F0, 0x808080, 'g', D_EXIT, 16, 0, (void*)"&Grid", NULL, (void *)ToggleGrid},
 { my_button_proc, MenuPosX-200, MenuPosY+100, 85, 20, 0xF0F0F0, 0x808080, 't', D_EXIT, 16, 0, (void*)"&TestClash", NULL, (void *)ClashTest},
 { my_button_proc, MenuPosX-100, MenuPosY+100, 85, 20, 0xF0F0F0, 0x808080,   0, D_EXIT, 16, 0, (void*)"BrightGrid", NULL, (void *)ToggleBriGrid},

//Palette Buttons & keys
 { my_button_proc, PalButtPosX+PalButtSizeX*0, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0xFFFFFF, C64Palette[0], 0, D_EXIT, 0, 0, (void*)"0&0", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*1, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[1], 0, D_EXIT, 1, 0, (void*)"0&1", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*2, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[2], 0, D_EXIT, 2, 0, (void*)"0&2", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*3, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[3], 0, D_EXIT, 3, 0, (void*)"0&3", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*4, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[4], 0, D_EXIT, 4, 0, (void*)"0&4", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*5, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[5], 0, D_EXIT, 5, 0, (void*)"0&5", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*6, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[6], 0, D_EXIT, 6, 0, (void*)"0&6", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*7, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[7], 0, D_EXIT, 7, 0, (void*)"0&7", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*8, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[8], 0, D_EXIT, 8, 0, (void*)"0&8", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*9, PalButtPosY,  PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[9], 0, D_EXIT, 9, 0, (void*)"0&9", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*10, PalButtPosY, PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[10], 0, D_EXIT, 10, 0, (void*)"0&A", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*11, PalButtPosY, PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[11], 0, D_EXIT, 11, 0, (void*)"0&B", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*12, PalButtPosY, PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[12], 0, D_EXIT, 12, 0, (void*)"0&C", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*13, PalButtPosY, PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[13], 0, D_EXIT, 13, 0, (void*)"0&D", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*14, PalButtPosY, PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[14], 0, D_EXIT, 14, 0, (void*)"0&E", NULL, (void*)ChangeCol },
 { my_button_proc, PalButtPosX+PalButtSizeX*15, PalButtPosY, PalButtSizeX, PalButtSizeY, 0x000000, C64Palette[15], 0, D_EXIT, 15, 0, (void*)"0&F", NULL, (void*)ChangeCol },
//GUI-control keys
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_ESC,   0,    (void *)Quit, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_PLUS_PAD, 0, (void *)Pluser, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_MINUS_PAD, 0, (void *)Minuser, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_SPACE, 0, (void *)Spacer, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_DEL, 0, (void *)Deleter, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_V, 0, (void *)Paster, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_X, 0, (void *)Cutter, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_HOME, 0, (void *)Homer, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_END, 0, (void *)Ender, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_PGUP, 0, (void *)PageUp, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_PGDN, 0, (void *)PageDn, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_LEFT, 0, (void *)Lefter, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_RIGHT, 0, (void *)Righter, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_UP, 0, (void *)Upper, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_DOWN, 0, (void *)Downer, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_Z, 0, (void *)Zeer, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_Y, 0, (void *)Yeer, NULL, NULL  },
//Palette keys 0..9
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_0, 0, (void *)SelCol0, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_1, 0, (void *)SelCol1, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_2, 0, (void *)SelCol2, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_3, 0, (void *)SelCol3, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_4, 0, (void *)SelCol4, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_5, 0, (void *)SelCol5, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_6, 0, (void *)SelCol6, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_7, 0, (void *)SelCol7, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_8, 0, (void *)SelCol8, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_9, 0, (void *)SelCol9, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_A, 0, (void *)SelColA, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_B, 0, (void *)SelColB, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_C, 0, (void *)SelColC, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_D, 0, (void *)SelColD, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_E, 0, (void *)SelColE, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F, 0, (void *)SelColF, NULL, NULL  },
//pencil-switcher keys
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F1, 0, (void *)F1er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F2, 0, (void *)F2er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F3, 0, (void *)F3er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F4, 0, (void *)F4er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F5, 0, (void *)F5er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F6, 0, (void *)F6er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F7, 0, (void *)F7er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F8, 0, (void *)F8er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F9, 0, (void *)F9er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F10, 0, (void *)F10er, NULL, NULL  },
 { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F11, 0, (void *)F11er, NULL, NULL  },
//stabilizer
 { d_yield_proc,        0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  },
 { NULL,          0,   0,   0,  0,  0,   0,   0,    0,      0,   0,   NULL,          NULL, NULL }
 };

//--------------------C64 'exe.prg'-export's native displayer code--------------
unsigned char bmpExe[]=
{
0x01,0x08,0x0b,0x08,0xdc,0x07,0x9e,0x32,0x30,0x36,0x34,0x00,0x00,0x00,0x00,0x00,
0x00,0x78,0xa9,0x35,0x85,0x01,0xa9,0x03,0x8d,0x00,0xdd,0x8d,0x11,0xd0,0xad,0x28,
0x2d,0x8d,0x20,0xd0,0x8d,0x21,0xd0,0x8d,0xab,0x09,0x20,0xb8,0x08,0x20,0x20,0x09,
0xa9,0x18,0x8d,0x18,0xd0,0xa9,0x3b,0x8d,0x11,0xd0,0xa9,0x08,0x8d,0x16,0xd0,0xa9,
0xf9,0xcd,0x12,0xd0,0xd0,0xfb,0xad,0x11,0xd0,0x09,0x20,0x29,0xf7,0x8d,0x11,0xd0,
0x29,0xf7,0x8d,0x11,0xd0,0xa9,0xff,0x8d,0x15,0xd0,0xa9,0xfa,0xcd,0x12,0xd0,0xd0,
0xfb,0xa2,0x09,0xca,0xd0,0xfd,0xad,0x11,0xd0,0x29,0xdf,0x8d,0x11,0xd0,0xa9,0x00,
0xcd,0x12,0xd0,0xd0,0xfb,0xa9,0x00,0x8d,0x21,0xd0,0xa9,0x09,0xcd,0x12,0xd0,0xd0,
0xfb,0xea,0xad,0x11,0xd0,0x09,0x08,0x29,0xdf,0x8d,0x11,0xd0,0xac,0xab,0x09,0xa2,
0x15,0xca,0xd0,0xfd,0xea,0x8c,0x21,0xd0,0xa9,0x00,0x8d,0x15,0xd0,0xad,0x12,0xd0,
0xd0,0xfb,0xa9,0x32,0xcd,0x12,0xd0,0xd0,0xfb,0xa2,0x0a,0xca,0xd0,0xfd,0xad,0x11,
0xd0,0x09,0x28,0x8d,0x11,0xd0,0x4c,0x3e,0x08,0xa0,0x00,0xb9,0x40,0x29,0x99,0x00,
0x04,0xb9,0x40,0x2a,0x99,0x00,0x05,0xb9,0x40,0x2b,0x99,0x00,0x06,0xb9,0x40,0x2c,
0x99,0x00,0x07,0xc8,0xd0,0xe5,0xa9,0x00,0xa2,0x29,0x85,0xfc,0x86,0xfd,0xa9,0x00,
0xa2,0x3f,0x85,0xfe,0x86,0xff,0xa0,0x00,0xb1,0xfc,0x91,0xfe,0xc8,0xd0,0xf9,0xc6,
0xfd,0xc6,0xff,0xa5,0xff,0xc9,0x20,0x10,0xef,0xa2,0x07,0x8a,0x0a,0xa8,0xa9,0x01,
0x99,0x01,0xd0,0xbd,0x8b,0x09,0x99,0x00,0xd0,0xa9,0x0f,0x9d,0x27,0xd0,0x8a,0x18,
0x69,0x30,0x9d,0xf8,0x07,0xca,0x10,0xe3,0xa9,0x00,0x8d,0x10,0xd0,0x8d,0xff,0x3f,
0x60,0xa0,0x00,0xa9,0x00,0x99,0x00,0x0c,0x99,0x00,0x0d,0xc8,0xd0,0xf5,0xa2,0x00,
0x86,0xfa,0xa6,0xfa,0xa9,0x00,0x8d,0x72,0x09,0x8d,0x73,0x09,0xbd,0xd8,0x09,0xf0,
0x4a,0x0a,0x2e,0x73,0x09,0x0a,0x2e,0x73,0x09,0x0a,0x2e,0x73,0x09,0x8d,0x72,0x09,
0xad,0x73,0x09,0x18,0x69,0xd8,0x8d,0x73,0x09,0xa9,0x00,0x18,0x7d,0x93,0x09,0x85,
0xfe,0xa0,0x0c,0xe0,0x0c,0x90,0x01,0xc8,0x84,0xff,0xa9,0x33,0x85,0x01,0xa0,0x00,
0xa2,0x00,0xbd,0x11,0x11,0x91,0xfe,0xc8,0xc8,0xc8,0xe8,0xe0,0x08,0xd0,0xf3,0xa9,
0x35,0x85,0x01,0xe6,0xfa,0xa6,0xfa,0xe0,0x18,0xd0,0xa7,0x60,0x50,0x68,0x80,0x98,
0xb0,0xc8,0xe0,0xf8,0x00,0x01,0x02,0x40,0x41,0x42,0x80,0x81,0x82,0xc0,0xc1,0xc2,
0x00,0x01,0x02,0x40,0x41,0x42,0x80,0x81,0x82,0xc0,0xc1,0xc2,0x00
};

//===================================================Main Function===============================================================
int main(int argc, char **argv) 
{	
    set_uformat(U_ASCII); //set character encoding to fix sized 8 bit ASCII characters
    //init window
	if (allegro_init()!= 0) return 1;
    LOCK_FUNCTION(CloseButton);
    set_close_button_callback(CloseButton);
	install_keyboard(); 
	install_mouse(); enable_hardware_cursor();
	install_timer();
    clear_keybuf();
    position_mouse_z(MagnifyFactor);

	set_window_title("HermIRES 1.0 - Hires Bitmap Editor for C64 - Hermit Software Hungary"); set_color_depth(24); 
    set_gfx_mode(GFX_AUTODETECT_WINDOWED, GuiSizeX, GuiSizeY, 0, 0);
    for (unsigned int i=0;i<16;i++) {BackPalette[C64Palette[i]]=i;} //copy C64Palette array to Backpalette associative map
    gui_bg_color = C64Palette[6]; gui_fg_color = C64Palette[14];
    DIALOG_PLAYER *player = init_dialog(OverDialog, -1); 
    update_dialog(player);
    show_mouse(screen);
    InitBmp(); //InitSpr();
    DrawGrid = create_bitmap(GridSizeX, GridSizeY);
    Preview = create_bitmap(FieldSizeX, FieldSizeY);
    ToolBar = create_bitmap(TBSizeX,TBSizeY);

    if (argc == 2) //checks whether not less/more than one command line option, which should be filename
    {//MIME to path - open the file from command line parameter
     for (int i=0;i < 480;i++) {path[i] = argv[1][i];}
     LoadFile();
    }

    BorderDrawer(); GridDrawer(GridCol); PrevDrawer(); PrevText(); DispSelCol(); DispToolBar(); //first display of bitmaps, etc.

    LOCK_VARIABLE(DrawGrid);
    LOCK_FUNCTION(GetControls);
    SetTimer();
//----------------------------------------------------------
    while (update_dialog(player)) //main loop of GUI dialog
    {
     if (CloseButtState) if (Quit()==D_CLOSE) return shutdown_dialog(player);
     if (PrevRightClick != 0) SaveBMP();
    }

    show_mouse(NULL);
    destroy_bitmap(DrawGrid);
    destroy_bitmap(Preview);
	return shutdown_dialog(player);
}
END_OF_MAIN();

//==========================================================Functions========================================================
//------------------------------Timed mouse event handler -------------------------------------------------------------------

void GetControls()  //timed function, handles keyboard and mouse events during dialog execution for the drawing part
{
 poll_mouse(); poll_keyboard(); CalcActChar();
 get_mouse_mickeys(&mickeyx,&mickeyy);
 textprintf_ex(screen,font,TxtInfoPosX, TxtInfoPosY+62,0x80FFFF,0x0000FF,"Grid-brighter-ratio: %1.2f", GridBrightness );
 textprintf_ex(screen,font,TxtInfoPosX, TxtInfoPosY+32,0x80FFFF,0x0000FF,"Pixel   X%.3d    Y%.3d", GetDataX(), GetDataY());

 ClashTimer--; 
 if (ClashTimer<0) { ClashTimer=ClashBlinkSpeed; BlinkOn=1; if(ClashTestMode){DetectClashes();GridDrawer(GridCol);} }
 else if (ClashTimer==ClashBlinkSpeed-(ClashBlinkSpeed/ClashBlinkRatio)) { BlinkOn=0; if(ClashTestMode){DetectClashes();GridDrawer(GridCol);} }

 if (mouse_b & 1)  //left mouse-button to draw selected Char-colour
 {
  if (CheckRange() == 0 ) //the cursor is in grid-window area
  {//UNDO point 1
   if(MButtBuf != mouse_b) StoreUndo();
   switch (DrawMode)
   {
    case 0: {DrawBmpData(GetDataX(),GetDataY(),DrawCol1);break;}
    case 1: case 2: case 3: case 5: ShapeDrawn=1;break;
    case 4: Filler(); break;
   }
   GridDrawer(GridCol);PrevDrawer();
  }
  if ((CheckRange() == 1) && (MButtBuf != mouse_b)) //mouse is in the preview area
  { 
   switch (PreviewZoom)  //switch preview-zoom values
   {
    case 1: PreviewZoom = 2;break;  
    case 2: PreviewZoom = 4;break; 
    case 4: default: PreviewZoom = 1;break; 
   }
   PrevDrawer(); PrevText();
  }
  if ((CheckRange() == 2) && (MButtBuf != mouse_b)) //mouse is in the ToolBar area
  {
   switch ((mouse_x - TBPosX)/TButtSize)
   {
    case 0: F11er();break;
    case 1: F9er(); break;
    case 2: {//increase diameter
             if (DrawMode == 0) {PenDiameter = (PenDiameter<32) ? PenDiameter+4:PenDiameter;} 
             else DrawMode=0;
             DispToolBar();break;}
    case 3: F5er(); break;
    case 4: F6er(); break;
    case 5: F7er(); break;
    case 6: F8er(); break;
    case 7: Undoer(); break;
   }
   //DispToolBar(); //included in key-subroutines
  }
  BuffPos();return;
 }
 else if (mouse_b & 2) //right mouse button to draw background colour
 {
  if (CheckRange() == 0 )
  {//UNDO point 2
   if(MButtBuf != mouse_b) StoreUndo();
   switch (DrawMode)
   {
    case 0: {DrawBmpData(GetDataX(),GetDataY(),DrawCol2);GridDrawer(GridCol); PrevDrawer();break;}
    case 1: case 2: case 3: {ShapeDrawn=1;ShapeRightButt=1;GridDrawer(GridCol); PrevDrawer();break;}
    case 4: {Filler();GridDrawer(GridCol); PrevDrawer();break;}
    case 5: 
    { //clipboard move checking
     ShapeDrawn=1;ShapeRightButt=1;
     if (mouse_x!=mouse_x_buf || mouse_y!=mouse_y_buf) GridDrawer(GridCol);
     mouse_x_buf=mouse_x; mouse_y_buf=mouse_y; 
     break; 
    }
   }
  }
  else if ((CheckRange() == 1) && (MButtBuf != mouse_b)) //the preview area
  {
   PrevRightClick++; 
  }
  else if ((CheckRange() == 2) && (MButtBuf != mouse_b)) //mouse is in the toolbar area
  {
   switch ((mouse_x - TBPosX)/TButtSize)
   { 
    case 0: F11er();break;
    case 1: F10er(); break;
    case 2: { //decrease diameter
             if (DrawMode==0) {PenDiameter = (PenDiameter>4) ? PenDiameter-4:PenDiameter;}
             else DrawMode=0;  
             DispToolBar();break;}
    case 7: Redoer(); break;
   }
  }
  BuffPos();return;
 } 
 else if (mouse_b & 4) //middle mouse-button to move zoomed area
 {
  if (CheckRange()==0)
  {
   GridInPosX-=GetDataX()-TempPosX;GridInPosY-=GetDataY()-TempPosY;
   SetBackRange();
   GridDrawer(GridCol);
   BuffPos();
  }
  return;
 } 
 else if ( mouse_z != WheelBuf ) //mouse wheel to increase/decrease magnification
 {
  Zoomer();
  BuffPos();return;
 }
//here is the point where no L/R mouse-button is pressed -------------------------------------------
 if (ShapeDrawn != 0) //a line/rectangle/circle should be drawn
 {
  ShapeDrawn = 0; //end of shapedrawing
  switch (DrawMode){
  case 1: LineShape(); break;
  case 2: RectDraw();break;
  case 3: CircleDraw(); break;
  case 5: CopyPasteSelection(); break;
  }
  ShapeRightButt=0;
 }

 BuffPos();  //this temp-updater happens in any wayz, both if mouse event happened or not
 TempStartX=GetDataX();TempStartY=GetDataY(); //temp x and y for starting position or center of line/circle/rect
 if (CheckRange() == 0) {mousestartx=mouse_x; mousestarty=mouse_y;}
}
END_OF_FUNCTION(GetControls)

//----------------------------------------------
void BuffPos() //store the changing mouse-event to detect singleclicks/wheel differences 
     {
      if (ShapeDrawn != 0)  //draw shape-preview on screen (in normal resolution)
      {
       if (CheckRange()==0)
       {
        switch (DrawMode)
        {
         case 1: {fastline(screen,mousestartx,mousestarty,mouse_x,mouse_y,0xFFFFFF); break;}
         case 2: {rect(screen,mousestartx,mousestarty,mouse_x,mouse_y,0xFFFFFF); break;}
         case 3: {int r= (int)(sqrt(pow(mouse_x-mousestartx,2)+pow(mouse_y-mousestarty,2))); 
                               circle(DrawGrid,mousestartx-GridPosX,mousestarty-GridPosY,r,0xFFFFFF);
                               blit(DrawGrid, screen, 0, 0, GridPosX, GridPosY, GridSizeX, GridSizeY);break;}
         case 5:
         { //clipboard copy/paste displays
          if (ShapeRightButt==0) rect(screen,mousestartx,mousestarty,mouse_x,mouse_y,0xFFFFFF);
          else DisplayClipBoard();
          break;
         }
        }
       }
      }
      //show_mouse(screen);
      WheelBuf = mouse_z; MButtBuf = mouse_b;
      TempPosX=GetDataX();TempPosY=GetDataY();  
     }


//-------------------------------------------------drawing, pixelling the graphic data------------------------------------
void DrawBmpData(int x, int y, unsigned char col)
     {
      if ((LinerSwitch != 0) && (MButtBuf == mouse_b)) {DrawBmpLine(TempPosX2,TempPosY2,x,y,col);}
      else {DrawChrPixel(x,y,col);}
      TempPosX2=GetDataX();TempPosY2=GetDataY(); 
     }

void LineShape() //draw line using mouse-click and release positions as endpoints
     {
      DrawBmpLine(TempStartX,TempStartY,GetDataX(),GetDataY(),(ShapeRightButt==0)?DrawCol1:DrawCol2);
      GridDrawer(GridCol);PrevDrawer();
     }

void RectDraw() //draw rectangle (normal/filled) using mouse-click and release positions as endpoints
     {
      int maxx=GetDataX(), maxy=GetDataY();
      if (RectFill == 0)
      {
       DrawBmpLine(TempStartX,TempStartY,maxx,TempStartY,(ShapeRightButt==0)?DrawCol1:DrawCol2);
       DrawBmpLine(TempStartX,TempStartY,TempStartX,maxy,(ShapeRightButt==0)?DrawCol1:DrawCol2);
       DrawBmpLine(TempStartX,maxy,maxx,maxy,(ShapeRightButt==0)?DrawCol1:DrawCol2);
       DrawBmpLine(maxx,TempStartY,maxx,maxy,(ShapeRightButt==0)?DrawCol1:DrawCol2);
      }
      else  //filled rectangle
      {
       if ((TempStartX <= maxx) && (TempStartY <= maxy))
       {
        for (int i=TempStartX;i<=maxx;i++) // += PenDiameter/4)
        {
         for (int j=TempStartY;j<=maxy;j++) // += PenDiameter/4)
         {
          DrawChrPixel(i,j,(ShapeRightButt==0)?DrawCol1:DrawCol2);
         }
        }
       }
       else if ((TempStartX > maxx) && (TempStartY <= maxy))
       {
        for (int i=TempStartX;i>=maxx;i--) // -= PenDiameter/4)
        {
         for (int j=TempStartY;j<=maxy;j++) // += PenDiameter/4)
         {
          DrawChrPixel(i,j,(ShapeRightButt==0)?DrawCol1:DrawCol2);
         }
        }
       }
       else if ((TempStartX <= maxx) && (TempStartY > maxy))
       {
        for (int i=TempStartX;i<=maxx;i++) // += PenDiameter/4)
        {
         for (int j=TempStartY;j>=maxy;j--) // -= PenDiameter/4)
         {
          DrawChrPixel(i,j,(ShapeRightButt==0)?DrawCol1:DrawCol2);
         }
        }
       }
       else
       {
        for (int i=TempStartX;i>=maxx;i--) // -= PenDiameter/4)
        {
         for (int j=TempStartY;j>=maxy;j--) // -= PenDiameter/4)
         {
          DrawChrPixel(i,j,(ShapeRightButt==0)?DrawCol1:DrawCol2);
         }
        }
       }
      }
      GridDrawer(GridCol);PrevDrawer();
     }

void Circler(int x, int y, int r, unsigned char col)
     {
      for (float angle = 0; angle < 360; angle+=0.5)
      {
       DrawChrPixel((int)(x+sin(angle*rad)*r+0.5), (int)(y+cos(angle*rad)*r+0.5), col);
      }
     }

void CircleDraw()
     {
      int maxx=GetDataX(), maxy=GetDataY();
      int radius = (int)(sqrt(pow(maxx-TempStartX,2)+pow(maxy-TempStartY,2))); 
      if (CircFill == 0)
      {
       Circler(TempStartX,TempStartY,radius,(ShapeRightButt==0)?DrawCol1:DrawCol2); 
      }
      else //filled circle
      {
       for (int i = 0; i<radius; i+= PenDiameter/4 )
       {
        Circler(TempStartX,TempStartY,i,(ShapeRightButt==0)?DrawCol1:DrawCol2);
       }
      }
      GridDrawer(GridCol);PrevDrawer();
     }

void Filler() //other bitmap with full zoom!!!
     {
      int zoomtemp = PreviewZoom;
      PreviewZoom=1;DrawPrv();
      if (RButt == 0) floodfill(Preview,GetDataX(),GetDataY(),C64Palette[DrawCol1]);
      else floodfill(Preview,GetDataX(),GetDataY(),C64Palette[DrawCol2]);
      ConvBackBMP(Preview);
      PreviewZoom = zoomtemp;
     }

void DrawBmpLine(int x1, int y1, int x2, int y2, unsigned char col)
     {
      int dy = y2 - y1;
      int dx = x2 - x1;
      float t = (float) 0.5;                      // offset for rounding

      DrawChrPixel(x1,y1,col);
      if (abs(dx) > abs(dy)) {          // slope < 1
            float m = (float) dy / (float) dx;      // compute slope
            t += y1;
            dx = (dx < 0) ? -1 : 1;
            m *= dx;
            while (x1 != x2) {
                x1 += dx;                           // step to next x value
                t += m;                             // add slope to y value
                DrawChrPixel(x1,(int) t,col);
            }
        } else {                                    // slope >= 1
            float m = (float) dx / (float) dy;      // compute slope
            t += x1;
            dy = (dy < 0) ? -1 : 1;
            m *= dy;
            while (y1 != y2) {
                y1 += dy;                           // step to next y value
                t += m;                             // add slope to x value
                DrawChrPixel((int) t,y1,col);
            }
        }
     }

void DrawChrPixel(int x, int y, unsigned char colour) //make the dot/plot/line, whatever in graphic data with coordinates & colour
     {
      unsigned int circorx,circory;
      for (int radius=0; radius < PenDiameter/2; radius++)
      {
       for (int angle = 0; angle < 360; angle += 2)
       {
        circorx = x+(int)(sin(angle*rad)*radius/2.2); circory = y+(int)(cos(angle*rad)*radius/2.2);
        if (0 <= circorx && circorx < FieldSizeX && 0 <= circory && circory < FieldSizeY) 
        {
         if (DitherMode==0) BmpData[circorx][circory]=colour; //mode 0 - no dither
         if (DitherMode==1) //mode 1 - dither even
         {
          if ((circorx%2)^(circory%2)==0) BmpData[circorx][circory]=colour;
          else {if (Dither2Col==1) BmpData[circorx][circory]=DrawCol2;} //two colour-mode for even dithering
         }
         if(DitherMode==2) //mode 2 - dither odd
         {
          if ((circorx%2)^(circory%2)==1) BmpData[circorx][circory]=colour;
          else {if (Dither2Col==1) BmpData[circorx][circory]=DrawCol2;} //two colour-mode for odd dithering
         }
        }
       }
      }
     }

//----------------------------------------------graph-positioning calculation routines---------------------------------------------
void Zoomer()  //zoom to the actual data-coordinate
     {
      if (CheckRange() != 0) {position_mouse_z(WheelBuf);return;}
      if (mouse_z < Mag_min) {position_mouse_z(Mag_min);GridInPosX=GridInPosY=0;}
      if (mouse_z > Mag_max) {position_mouse_z(Mag_max);}
      int PrevDataX=GetDataX(), PrevDataY=GetDataY();
      MagnifyFactor = mouse_z;
      int DeltaX = GetDataX()-PrevDataX, DeltaY = GetDataY()-PrevDataY; GridInPosX-=DeltaX; GridInPosY-=DeltaY;
      SetBackRange();
      GridDrawer(GridCol);
     }

void CalcActChar()
     {
      if (CheckRange() != 0) return;
      int x = GetDataX()/ChrSizeX;
      int y = GetDataY()/ChrSizeY;
      ActChar = x + y * NoOfChrX;
      if (ActChrBuf != ActChar) 
      {
       TextDispl();
       if (PreviewZoom != 1) PrevDrawer();
      }
      ActChrBuf = ActChar;
     }

int CheckRange() //checks whether the mouse is in the gridfield/datafield - avoid border-effects
    {            //also checks whether the mouse is in the preview-screen or not
     if ((GridPosX < mouse_x) && (mouse_x < GridPosX+GridSizeX) && (GridPosY < mouse_y) && (mouse_y < GridPosY+GridSizeY) && 
         (mouse_x < GridPosX+(FieldSizeX-GridInPosX)*MagnifyFactor) && (mouse_y < GridPosY+(FieldSizeY-GridInPosY)*MagnifyFactor)) return 0;
     if ((PreviewPosX <mouse_x) && (mouse_x < (PreviewPosX+FieldSizeX)) && 
        (PreviewPosY < mouse_y) && (mouse_y < (PreviewPosY+FieldSizeY))) return 1;
     if ((TBPosX <mouse_x) && (mouse_x < (TBPosX+TBSizeX)) && 
        (TBPosY < mouse_y) && (mouse_y < (TBPosY+TBSizeY))) return 2;
     return -1;
    }

void SetBackRange()
     {
      if (GridInPosX < 0) GridInPosX=0;
      if (GridInPosY < 0) GridInPosY=0;
      if ( (FieldSizeX-GridInPosX)*MagnifyFactor < GridSizeX ) GridInPosX=FieldSizeX-GridSizeX/MagnifyFactor;
      if ( (FieldSizeY-GridInPosY)*MagnifyFactor < GridSizeY ) GridInPosY=FieldSizeY-GridSizeY/MagnifyFactor;
      if (MagnifyFactor == Mag_min) GridInPosX=GridInPosY=0;
     }

int GetDataX() //calculate data pointer from window-relative mouse x-coordinate, handle outranges
    {
     return (mouse_x-1-GridPosX)/MagnifyFactor+GridInPosX;
    }

int GetDataY() //calculate data pointer from window-relative mouse y-coordinate, handle outranges
    {
     return (mouse_y-1-GridPosY)/MagnifyFactor+GridInPosY;
    }
//---------------------------------------------------PIXELDRAWER/GRAPHDRAWER functions for display--------------------------------
unsigned int ClashCol (unsigned int col)
{
 int r=getr(col), g=getg(col), b=getb(col);
 if (BlinkOn) {r=(int)(r*GridBrightness+100); g=(int)(g*GridBrightness+50); b=(int)(b*GridBrightness+50); } 
 else { r=(int)(r*GridBrightness); g=(int)(g*GridBrightness); b=(int)(b*GridBrightness); }
 if (r<0) r=0; if(g<0) g=0; if(b<0) b=0;
 if (r>255) r=255; if(g>255) g=255; if(b>255) b=255;
 return makecol(r,g,b);
}

unsigned int BrightComp(unsigned int col)
{
 int r=getr(col), g=getg(col), b=getb(col);
 r=(int)(r*GridBrightness); g=(int)(g*GridBrightness); b=(int)(b*GridBrightness);
 if (r>255) r=255; if(g>255) g=255; if(b>255) b=255;
 return makecol(r,g,b);
}

 //float hue, saturation, value;
 //rgb_to_hsv(r, g, b, &hue, &saturation, &value);
 //hsv_to_rgb(hue, saturation, value, &r, &g, &b);

void DrawGridPixel(int X, int Y, int Col, int chkChar)
     {
      if ((ClashCnt[chkChar] > 2) && ClashTestMode)
      {
       if (BlinkOn) rect(DrawGrid, X*MagnifyFactor, Y*MagnifyFactor, X*MagnifyFactor+MagnifyFactor, Y*MagnifyFactor+MagnifyFactor, 0x804020);
       rectfill(DrawGrid, X*MagnifyFactor+1, Y*MagnifyFactor+1, 
       X*MagnifyFactor+MagnifyFactor-1, Y*MagnifyFactor+MagnifyFactor-1, ClashCol(C64Palette[Col]) );    
      }
      else
      {
       rectfill(DrawGrid, X*MagnifyFactor+EnableGrid, Y*MagnifyFactor+EnableGrid, 
       X*MagnifyFactor+MagnifyFactor-1, Y*MagnifyFactor+MagnifyFactor-1, BrightComp(C64Palette[Col]) );    
      }
     }

void DrawPrevPixel (int X, int Y, int Col)
     {
      putpixel(Preview,X,Y,C64Palette[Col]);
     }

void BmpDrawer() //draws the magnified Bitmap
     {
      int chkChar;
      for (int i = 0;i<(GridSizeX/MagnifyFactor+1) && i<(FieldSizeX-GridInPosX+1);i++)
      {
       for (int j = 0;j<(GridSizeY/MagnifyFactor+1) && j<(FieldSizeY-GridInPosY+1);j++)
       {
        chkChar=(i+GridInPosX)/ChrSizeX+((j+GridInPosY)/ChrSizeY)*NoOfChrX;
        //if (BmpData[i+GridInPosX][j+GridInPosY]<16)
        //{
         DrawGridPixel(i,j,BmpData[i+GridInPosX][j+GridInPosY],chkChar);
        //}
        //else
        //{
        // DrawGridPixel(i,j,BGCol,chkChar);
        //}
       }
      }
     }

void TextDispl()
     {
      int x = GetDataX()/ChrSizeX;
      int y = GetDataY()/ChrSizeY;
      ActQuad=(ActChar/8)*2+(ActChar-(ActChar/4)*4)/2+1;
      textprintf_ex(screen,font,TxtInfoPosX,TxtInfoPosY+42,0x80FFFF,0x0000FF,"Char  X%.2d Y%.2d Memo:$%.3X", x, y, ActChar);

      //textprintf_ex(screen,font, MenuPosX-370, MenuPosY+90, 0x80FFFF,0x000000,(WideMode)?"1280":"960 ");
      textprintf_ex(screen,font, MenuPosX-270, MenuPosY+90, 0x80FFFF,0x000000,(EnableGrid)?"ON ":"OFF");
      textprintf_ex(screen,font, MenuPosX-170, MenuPosY+90, 0x80FFFF,0x000000,(ClashTestMode)?"ON":"  ");
      textprintf_ex(screen,font, MenuPosX-70, MenuPosY+90, 0x80FFFF,0x000000,(EnableBriGrid)?"ON ":"OFF");
     }

void PrevText()
     {
      textprintf_ex(screen,font,TxtInfoPosX+88,TxtInfoPosY+52,0x80FFFF,0x0000FF," Preview: %dX", PreviewZoom);
     }

void BorderDrawer()
     {
      rectfill(screen, GridPosX-BorderColWidth,GridPosY-BorderColWidth,GridPosX+GridSizeX+BorderColWidth,GridPosY+GridSizeY+BorderColWidth,C64Palette[BorderCol]);
      rectfill(screen, PreviewPosX-BorderColWidth,PreviewPosY-BorderColWidth,PreviewPosX+FieldSizeX+BorderColWidth,PreviewPosY+FieldSizeY+BorderColWidth,C64Palette[BorderCol]);
     }

void GridDrawer(int GridCol)  //function to draw magnify-grid
     {
      int GridWidth = MagnifyFactor*FieldSizeX;
      int GridHeight = MagnifyFactor*FieldSizeY;
      GridBrightness=(EnableGrid && EnableBriGrid) ? (float)MagnifyFactor/(float)(MagnifyFactor-1) : 1.00; 
      for(int i=0,j=0; i<(GridWidth+1); i+=MagnifyFactor,j++)
      {
       fastline(DrawGrid,i,0,i,GridHeight, ((j+GridInPosX)%ChrSizeX) ? GridCol : ChrRangeCol );
      }
      for(int i=0,j=0; i<(GridHeight+1); i+=MagnifyFactor,j++)
      {
       fastline(DrawGrid,0,i,GridWidth,i,((j+GridInPosY)%ChrSizeY == 0) ? ChrRangeCol:GridCol);
      }
      BmpDrawer();
      textprintf_ex(screen,font,TxtInfoPosX,TxtInfoPosY+52,0x80FFFF,0x0000FF,"  Zoom: %dX ", MagnifyFactor);
      blit(DrawGrid, screen, 0, 0, GridPosX, GridPosY, GridSizeX, GridSizeY);
     }

void DrawPrv() //draw preview bitmap without blit to screen (invisible change) - used by fill routine
     {
      int ZoomCordX, ZoomCordY, ZoomPixCol, PrevStartX, PrevStartY;
      int xmul=ChrSizeX*NoOfChrX/PreviewZoom, ymul=ChrSizeY*NoOfChrY/PreviewZoom;
      if (PreviewZoom==1) PrevStartX=PrevStartY=0;
      else
      {
       PrevStartX = SelChrPosX(ActChar)-xmul/2;
       PrevStartY = SelChrPosY(ActChar)-ymul/2;
       if (PrevStartX < 0) PrevStartX=0;
       if (PrevStartY < 0) PrevStartY=0;
       if (PrevStartX >= FieldSizeX-xmul) PrevStartX=FieldSizeX-xmul;
       if (PrevStartY >= FieldSizeY-ymul) PrevStartY=FieldSizeY-ymul;
      }
      for (int i = 0;i<FieldSizeX;i++)
      {
       for (int j = 0;j<FieldSizeY;j++)
       {
        ZoomCordX = i/PreviewZoom+PrevStartX ; ZoomCordY = j/PreviewZoom+PrevStartY ;
        ZoomPixCol = BmpData[ZoomCordX][ZoomCordY];
        //if (ZoomPixCol<16)
        //{
         DrawPrevPixel(i,j, ZoomPixCol);
        //}
        //else
        //{
        // DrawPrevPixel(i,j, BGCol);
        //}
       }
      }
     }

void PrevDrawer() //function to draw preview box - using variable 'PreviewZoom' for magnification factoring
     {
      DrawPrv();
      blit(Preview, screen, 0, 0, PreviewPosX, PreviewPosY, FieldSizeX+1, FieldSizeY+1);
     }

void DispToolBar()
     {
      clear_to_color(ToolBar, TBCol);
      for (int i=0;i<NoOfButt;i++){rect(ToolBar, i*TButtSize, 0, (i+1)*TButtSize, TBSizeY-1, 0xF0F0F0);} //draw graph-button borders
      rect (ToolBar, 4, 4, TButtSize-4,TButtSize-4, (DrawMode==5) ? TBSelCol:0x505050);  //button0 - selection tool
       rect(ToolBar,TButtSize/2-3,2,TButtSize/2+3,TButtSize-2,TBCol); rect(ToolBar,2,TButtSize/2-3,TButtSize-2,TButtSize/2+3,TBCol);
      if (Dither2Col == 1) rectfill(ToolBar,TButtSize+8,8,TButtSize+16,16,TBSelCol);
      if (DitherMode == 0) rectfill(ToolBar,TButtSize+8,8,TButtSize+14,14,TBFGCol); //button 1 - dithermode
      else if (DitherMode == 1) {rectfill(ToolBar,TButtSize+8,8,TButtSize+11,11,TBFGCol);rectfill(ToolBar,TButtSize+12,12,TButtSize+15,15,TBFGCol);}
      else if (DitherMode == 2) {rectfill(ToolBar,TButtSize+12,8,TButtSize+15,11,TBFGCol);rectfill(ToolBar,TButtSize+8,12,TButtSize+11,15,TBFGCol);}                            
      circlefill(ToolBar,(int)(TButtSize*2.5),TButtSize/2,PenDiameter/4,(DrawMode==0)?TBSelCol:TBFGCol); //button 2 - dot mode (0)
      fastline(ToolBar,TButtSize*3+5,5,TButtSize*4-5,TBSizeY-5,(DrawMode==1)?TBSelCol:TBFGCol); //button 3 - line mode (1)
      if (RectFill == 0) rect(ToolBar,TButtSize*4+5,5,TButtSize*5-5,TBSizeY-5,(DrawMode==2)?TBSelCol:TBFGCol); //button 4 - rectangle modes (2)
         else rectfill(ToolBar,TButtSize*4+5,5,TButtSize*5-5,TBSizeY-5,(DrawMode==2)?TBSelCol:TBFGCol);
      if (CircFill == 0) circle(ToolBar,(int)(TButtSize*5.5),TBSizeY/2,TButtSize/2-4,(DrawMode==3)?TBSelCol:TBFGCol); //button 5 - circle modes (3)
         else circlefill(ToolBar,(int)(TButtSize*5.5),TBSizeY/2,TButtSize/2-4,(DrawMode==3)?TBSelCol:TBFGCol);
      circlefill(ToolBar,(int)(TButtSize*6.5),(int)(TBSizeY*2.5/4),4,(DrawMode==4)?TBSelCol:TBFGCol); //button6 - fill mode (4)
      triangle(ToolBar,TButtSize*6+9,TBSizeY/2,TButtSize*6+15,TBSizeY/2,(int)(TButtSize*6.5),4,(DrawMode==4)?TBSelCol:TBFGCol);
      triangle(ToolBar,TButtSize*7+5,TBSizeY/2,TButtSize*7+10,TBSizeY/2-5,TButtSize*7+10,TBSizeY/2+5,TBFGCol); //undo (left click)
      triangle(ToolBar,TButtSize*7+17,TBSizeY/2,TButtSize*7+12,TBSizeY/2-5,TButtSize*7+12,TBSizeY/2+5,TBFGCol); //redo (right click)
      blit(ToolBar, screen, 0, 0, TBPosX, TBPosY, TBSizeX, TBSizeY);
     }

void InitBmp()  //fills background colour
     {
      for (int i = 0;i<FieldSizeX+8;i++)
      {
       for (int j = 0;j<FieldSizeY+8;j++)
       {
        BmpData[i][j] = (i>=FieldSizeX || j>=FieldSizeY) ? 0:DrawCol2; //16;
       }
      }
      for (int i=0; i<1040; i++) ClashCnt[i]=0;
     }

void DispSelCol()
     {
      textprintf_ex(screen,font,TxtInfoPosX, TxtInfoPosY+00,0x80FFFF,0x0000FF," Drawing Colour 1: %s",Hex[DrawCol1]);
      textprintf_ex(screen,font,TxtInfoPosX, TxtInfoPosY+10,0x80FFFF,0x0000FF," Drawing Colour 2: %s",Hex[DrawCol2]);
      textprintf_ex(screen,font,TxtInfoPosX, TxtInfoPosY+20,0x80FFFF,0x0000FF," C64 Border Color: %s",Hex[BorderCol]);
      textprintf_ex(screen,font,TxtInfoPosX+165, TxtInfoPosY+00,0,C64Palette[DrawCol1]," ");
      textprintf_ex(screen,font,TxtInfoPosX+165, TxtInfoPosY+10,0,C64Palette[DrawCol2]," ");
      textprintf_ex(screen,font,TxtInfoPosX+165, TxtInfoPosY+20,0,C64Palette[BorderCol]," ");
     }

//---------------------------------------------------------------------------------------------------------
//------------------GUI-dialog event handling callback routines--------------------------------------------
int SelCol0()    //routine for all hexa keys, allegro GUI dialog sucks! - or I'm too new to it, and cannot pass arguments :(
 {               //there were other ways to reduce code, but I was too lazy, and wanted a homogene interface handler with GUI
  SelCol=0; return ChangeCol();
 }

int SelCol1()
 {
  SelCol=1; return ChangeCol();
 }

int SelCol2()
 {
  SelCol=2; return ChangeCol();
 }

int SelCol3()
 {
  SelCol=3; return ChangeCol();
 }

int SelCol4()
 {
  SelCol=4; return ChangeCol();
 }

int SelCol5()
 {
  SelCol=5; return ChangeCol();
 }

int SelCol6()
 {
  SelCol=6; return ChangeCol();
 }

int SelCol7()
 {
  SelCol=7; return ChangeCol();
 }

int SelCol8()
 {
  SelCol=8; return ChangeCol();
 }

int SelCol9()
 {
  SelCol=9; return ChangeCol();
 }

int SelColA()
 {
  SelCol=10; return ChangeCol();
 }

int SelColB()
 {
  if (key_shifts & KB_CTRL_FLAG)
  {
   return SaveBMP();
  }
  else
  {
  SelCol=11; return ChangeCol();
  }
 }

int SelColC()
 {
  if (key_shifts & KB_CTRL_FLAG)
  {
   return CopyChar();
  }
  else
  {
  SelCol=12; return ChangeCol();
  }
 }

int SelColD()
 {
  SelCol=13; return ChangeCol();
 }

int SelColE()
 {
  SelCol=14; return ChangeCol();
 }

int SelColF()
 {
  SelCol=15; return ChangeCol();
 }

int ChangeFullCol()
    {
     //UNDO POINT
     StoreUndo();     
     for (int i=0;i<FieldSizeX;i++)
     {
      for (int j=0;j<FieldSizeY;j++)
      {
       if (BmpData[i][j] == DrawCol1) BmpData[i][j]=SelCol;
      }
     }
    }

int ChangeCol()
    {
     remove_timer();
     if (RButt==2) //middle button?
     { //select Border-colour
       //UNDO POINT
       //StoreUndo();
       BorderCol=SelCol;BorderDrawer();GridDrawer(GridCol);PrevDrawer();DispSelCol(); 
     }
     else if (RButt==1) { DrawCol2=SelCol;DispSelCol(); }
     else if (key_shifts & KB_CTRL_FLAG) {ChangeFullCol();GridDrawer(GridCol);PrevDrawer();DrawCol1=SelCol;DispSelCol();} //change a colour entirely
     else if (key_shifts & KB_SHIFT_FLAG) {DrawCol2=SelCol;DispSelCol();}
     else {DrawCol1=SelCol;DispSelCol();} //no shift keys, normal colour-selection
     SetTimer();
     return D_O_K;
    }

//==========================================Custom GUI button handlers======================================================

void CloseButton()
     {
      CloseButtState=TRUE;
     }
END_OF_FUNCTION(CloseButton);

int Quit(void)
{
   remove_timer();
   if (alert("Really Quit?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1)
       return D_CLOSE;
   else
      {CloseButtState=FALSE;position_mouse_z(WheelBuf); SetTimer(); return D_O_K;}
}

/* A custom dialog procedure, derived from d_button_proc. It intercepts
 * the D_CLOSE return of d_button_proc, and calls the function in dp3.
 */
int my_button_proc(int msg, DIALOG *d, int c)
{
   if (mouse_b & 2) RButt=1; else if (mouse_b & 4) RButt=2; else RButt=0;
   int ret = d_button_proc(msg, d, c);
   if (ret == D_CLOSE && d->dp3)
     {
      SelCol=d->d1; //signs color selection
      return ((int (*)(void))d->dp3)();
     }
   return ret;
}

void SetTimer()
     {
      install_int_ex(GetControls, MSEC_TO_TIMER(ControlTimer)); //timer for controls
     }
//-------------------------------------------------------------------------------------------------------------
int SelChrPosX(int Chrnum)
  {
   return (Chrnum - (int)(Chrnum/NoOfChrX)*NoOfChrX ) * ChrSizeX; 
  }

int SelChrPosY(int Chrnum)
  {
   return (Chrnum/NoOfChrX) * ChrSizeY;
  }

void SelectChar(int Chrnum) //zooms to the needed Char
     {/*
      //ActChar = Chrnum; //global variable saying which Char is selected (cursor above)
       
      //GridInPosX= SelChrPosX(Chrnum);
      //GridInPosY= SelChrPosY(Chrnum);
      WheelBuf=MagnifyFactor=Mag_max; position_mouse_z(MagnifyFactor);
      GridDrawer(GridCol);PrevDrawer(); TextDispl();*/
     }

int Pluser()  //numpad plus key handler
 {
  remove_timer(); //prevent call-interference with timer
  //if (key_shifts & KB_SHIFT_FLAG)
  //{
   if (CheckRange()==0) {WheelBuf++; position_mouse_z(WheelBuf);}
   Zoomer();
  //}
  /*else
  {
   if (key_shifts & KB_CTRL_FLAG)
   {
    if(GridInPosX == 0) GridInPosX = ChrSizeX*2; 
    else 
    {
     GridInPosX=0;
     if(GridInPosY == 0) GridInPosY = ChrSizeY*2; else GridInPosY=0;
    }
    WheelBuf=MagnifyFactor = 8; position_mouse_z(WheelBuf);GridDrawer(GridCol);PrevDrawer();
   }
   else
   {
    if (ActChar < 15 ) ActChar++;
    SelectChar(ActChar);
   }
  }*/
  SetTimer();
  return D_O_K;
 }

int Minuser()  //numpad minus key handler
 {
  remove_timer(); //prevent call-interference with timer
  //if (key_shifts & KB_SHIFT_FLAG)
  //{
   if (CheckRange()==0) {WheelBuf--; position_mouse_z(WheelBuf);}
   Zoomer();
  //}
  /*else
  {
   if (key_shifts & KB_CTRL_FLAG)
   {
    if(GridInPosX == 0) 
    {
     GridInPosX = ChrSizeX*2; 
     if(GridInPosY == 0) GridInPosY = ChrSizeY*2; else GridInPosY=0;
    }
    else GridInPosX=0;
    WheelBuf=MagnifyFactor = 8; position_mouse_z(WheelBuf);GridDrawer(GridCol);PrevDrawer();
   }
   else
   {
    if (ActChar > 0 ) ActChar--;
    SelectChar(ActChar);
   }
  }*/
  SetTimer();
  return D_O_K;
 }

int Spacer()   //SPACE key handler
 {
  remove_timer();
  if (key_shifts & (KB_SHIFT_FLAG | KB_CTRL_FLAG))
  {
   if (CheckRange() == 0)
   {
    if (MagnifyFactor == Mag_max/2) WheelBuf=Mag_min; else WheelBuf=Mag_max/2;
   }
  }
  else
  {
   if (CheckRange() == 0)
   {
    if (MagnifyFactor == Mag_max) WheelBuf=Mag_min; else WheelBuf=Mag_max;            //{SelectChar(ActChar);}
   }
  }
  position_mouse_z(WheelBuf); Zoomer();
  SetTimer();  
  return D_O_K;
 } 

int Deleter()   //delete one Char
 {
  remove_timer();
  //UNDO POINT
  StoreUndo();
  int ipos = SelChrPosX(ActChar), jpos = SelChrPosY(ActChar);
  for (int i=ipos; i<(ipos+ChrSizeX); i++)
  {
   for (int j=jpos; j<(jpos+ChrSizeY); j++)
   {
    BmpData[i][j]=(key_shifts & KB_SHIFT_FLAG) ? DrawCol1:DrawCol2;
   }
  }
  GridDrawer(GridCol);PrevDrawer();
  SetTimer();
  return D_O_K;
 }

int CopyChar()    //copies actual Char to buffer 
 {
  remove_timer();
  int ipos = SelChrPosX(ActChar), jpos = SelChrPosY(ActChar), chkifempty=0;
  /*for (int i=ipos; i<(ipos+ChrSizeX); i++)
  {
   for (int j=jpos; j<(jpos+ChrSizeY); j++)
   {
    if (BmpData[i][j] < 16) chkifempty++;
   }
  }
  if (chkifempty > 0) 
  {*/
   for (int i=ipos; i<(ipos+ChrSizeX); i++)
   {
    for (int j=jpos; j<(jpos+ChrSizeY); j++)
    {
     CopyBuffer[(i-ipos)][(j-jpos)] = BmpData[i][j];
    }
   }
  //}
  SetTimer();
  return D_O_K;
 }

int Paster()    //copies actual Char to buffer 
 {
  if (key_shifts & KB_CTRL_FLAG)
  {
   remove_timer();
   //UNDO POINT
   StoreUndo();
   int ipos = SelChrPosX(ActChar), jpos = SelChrPosY(ActChar);
   for (int i=ipos; i<(ipos+ChrSizeX); i++)
   {
    for (int j=jpos; j<(jpos+ChrSizeY); j++)
    {
     BmpData[i][j] = CopyBuffer[(i-ipos)][(j-jpos)];
    }
   }
   GridDrawer(GridCol);PrevDrawer();
   SetTimer();
  }
  return D_O_K;
 }

int Cutter() //cut Char and send deleted data to buffer
    {
     CopyChar(); Deleter();
     return D_O_K;
    }

int Homer()
    {
     remove_timer();
     GridInPosX=0;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int Ender()
    {
     remove_timer();
     GridInPosX=FieldSizeX-GridSizeX/MagnifyFactor;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int PageUp()
    {
     remove_timer();
     GridInPosY=0;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int PageDn()
    {
     remove_timer();
     GridInPosY=FieldSizeY-GridSizeY/MagnifyFactor;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int Lefter()
    {
     remove_timer();
     if (GridInPosX>8) GridInPosX-=8; else GridInPosX=0;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int Righter()
    {
     remove_timer();
     int maxGrPosX=FieldSizeX-GridSizeX/MagnifyFactor;
     if (GridInPosX<maxGrPosX-8) GridInPosX+=8; else GridInPosX=maxGrPosX;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int Upper()
    {
     remove_timer();
     if (GridInPosY>8) GridInPosY-=8; else GridInPosY=0;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

int Downer()
    {
     remove_timer();
     int maxGrPosY=FieldSizeY-GridSizeY/MagnifyFactor;
     if (GridInPosY<maxGrPosY-8) GridInPosY+=8; else GridInPosY=maxGrPosY;
     GridDrawer(GridCol);
     SetTimer();
     return D_O_K;
    }

void CopyPasteSelection()
    {
     if (ShapeRightButt==0)
     { //copy selection to copy-buffer
      int maxx=GetDataX(), maxy=GetDataY();
      ClipBoardSizeX=0; //init
      {
       if ((TempStartX <= maxx) && (TempStartY <= maxy))
       {
        for (int i=TempStartX;i<=maxx&&i<FieldSizeX;i++,ClipBoardSizeX++)
        {
         ClipBoardSizeY=0; //init
         for (int j=TempStartY;j<=maxy&&j<FieldSizeY;j++,ClipBoardSizeY++)
         {
          BmpClipBoard[i-TempStartX][j-TempStartY]=BmpData[i][j];
         }
        }
       }
       else if ((TempStartX > maxx) && (TempStartY <= maxy))
       {
        for (int i=TempStartX;i>=maxx&&i>=0;i--,ClipBoardSizeX++)
        {
         ClipBoardSizeY=0; //init
         for (int j=TempStartY;j<=maxy&&j<FieldSizeX;j++,ClipBoardSizeY++)
         {
          BmpClipBoard[TempStartX-i][j-TempStartY]=BmpData[i][j];
         }
        }
       }
       else if ((TempStartX <= maxx) && (TempStartY > maxy))
       {
        for (int i=TempStartX;i<=maxx&&i<FieldSizeX;i++,ClipBoardSizeX++) 
        {
         ClipBoardSizeY=0; //init
         for (int j=TempStartY;j>=maxy&&j>=0;j--,ClipBoardSizeY++)
         {
          BmpClipBoard[i-TempStartX][TempStartY-j]=BmpData[i][j];
         }
        }
       }
       else
       {
        for (int i=TempStartX;i>=maxx&&i>=0;i--,ClipBoardSizeX++) 
        {
         ClipBoardSizeY=0; //init
         for (int j=TempStartY;j>=maxy&&j>=0;j--,ClipBoardSizeY++) 
         {
          BmpClipBoard[TempStartX-i][TempStartY-j]=BmpData[i][j];
         }
        }
       }
      }
     }
     else
     { //paste selection to the position where right mouse-button is released
      int PastePosX=GetDataX(), PastePosY=GetDataY();
      if (PastePosX<0 || PastePosY<0) return; //to be on the safe side
      for (int i=0;(i<ClipBoardSizeX)&&(PastePosX+i<FieldSizeX);i++)
        {
         for (int j=0;(j<ClipBoardSizeY)&&(PastePosY+j<FieldSizeY);j++)
         {
          BmpData[i+PastePosX][j+PastePosY] = BmpClipBoard[i][j];
         }
        }
     }
     GridDrawer(GridCol); PrevDrawer();
    }

void DisplayClipBoard()
     {
      int GridMpX=mouse_x-GridPosX-1, GridMpY=mouse_y-GridPosY-1; 
      int disposX=GridMpX-(GridMpX % MagnifyFactor)+GridPosX, disposY= GridMpY-(GridMpY % MagnifyFactor)+GridPosY, pixX, pixY;
      for (int i=0;i<ClipBoardSizeX;i++)
      {
       for (int j=0;j<ClipBoardSizeY;j++)
       {
        pixX=disposX+i*MagnifyFactor; pixY=disposY+j*MagnifyFactor;
        if (pixX>=GridPosX+GridSizeX || pixY>=GridPosY+GridSizeY) continue;
        else rectfill(screen,pixX ,pixY , (pixX+MagnifyFactor>=GridPosX+GridSizeX-1) ? GridPosX+GridSizeX-1 : pixX+MagnifyFactor-1, 
              (pixY+MagnifyFactor>=GridPosY+GridSizeY-1) ? GridPosY+GridSizeY-1 : pixY+MagnifyFactor-1, C64Palette[ BmpClipBoard[i][j] ] );
       }
      }
     }

//=================================================GUI button procedures============================
int ClearBmp()
    {
     remove_timer();
     if (alert("Really Clear the whole Bitmap?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1)
       {
        //UNDO POINT
        StoreUndo();
        InitBmp();
        GridDrawer(GridCol);PrevDrawer();
        SetTimer();
        return D_O_K;
       }
     else
      {CloseButtState=FALSE;position_mouse_z(WheelBuf); SetTimer(); return D_O_K;}
    }

int Help(void)
    {
     remove_timer();
     alert("mouse->left:draw, right:rubber, middle:move, scroll:zoom",
           "0..9 A..F - select drawcolour, with SHIFT - drawcolour 2",
           "SPACE-zoomswitch, F1..F4-pendiameter, F5..F8-drawingmode",
           "Check out the README.txt for more details!", 0, 0, 0); 
     SetTimer();
     return D_O_K;
    }
//--------------------------------------FILE OPERATIONS----------------------------------
/*void ConvertBackChr() //convert a Char-layer back from prg-file to hbm graph-data
     {
      unsigned char colour = (ConvChr[0][8] & 0x0f) , Charnum = ((ConvChr[0][8] & 0xf0)/16);
      int ipos=SelChrPosX(Charnum), jpos=SelChrPosY(Charnum), Chrx, Chry, mempos, bytepos;
      for (int i=ipos; i<(ipos+ChrSizeX); i++)
      {
       for (int j=jpos; j<(jpos+ChrSizeY); j++)
       {
        Chrx= i-ipos; Chry= j-jpos;
        mempos = (Chry)*3 + (Chrx)/8; bytepos = Chrx-(Chrx/8)*8;
        if ((ConvChr[0][mempos] & pow2[bytepos]) != 0)
        { //set the hbm Char-pixel for the actual colour
         BmpData[i][j] = colour;
        }
       }
      }
     }*/

void LoadPrg(char *filepath)   //write a new binary file with the graphic-data  
     {
      int i,j,xrough,yrough,yfine, prgaddr=0; //counts the read Chars, whenever reaches max., protects the loader
      InitBmp();   
      PACKFILE *PRGFile = pack_fopen(filepath,"r");
      pack_fread(&prgaddr,2,PRGFile); //copy the unneeded loadaddress to a yet unused space, will be overwritten soon
      if (prgaddr==0x801) pack_fseek(PRGFile,EXEdataAdd-BASICstartAdd);
      pack_fread(&C64bmp,0x1F40,PRGFile);  pack_fread(&ColorMap,0x03E8,PRGFile);  pack_fread(&BorderCol,1,PRGFile);
      for (i=0;i<0x1F40;i++)
      {
       for (j=0;j<8;j++)
       {
        yrough=(i/320)*8; yfine=(i-yrough*40)%8; xrough=(i-yrough*40)-yfine;
        BmpData[xrough+j][yrough+yfine] = ((C64bmp[i]&pow2[j])==0) ? ColorMap[i/8]&0x0F : ColorMap[i/8]/16;
       }
      }
      pack_fclose(PRGFile);
     } 

int Load()   //.prg or .hbm (.bmp/png ?) loader
    {
     int i;
     remove_timer();
     if (file_select_ex("Select File (.hbm/.prg/.bmp) to open",path,"hbm;prg;bmp",480,400,400) != 0)
     {
      LoadFile();
     }
     WheelBuf=MagnifyFactor=Mag_min; position_mouse_z(WheelBuf); GridInPosX=GridInPosY=0;
     BorderDrawer(); GridDrawer(GridCol); PrevDrawer(); DispSelCol();
     for (i=0;i<1000;i++) ClashCnt[i]=0; //init
     SetTimer();
     return D_O_K;
    }

void LoadFile()
     {
      int i,j;
      fix_filename_slashes(path);
      if (exists(path))
      {
       //UNDO POINT
       StoreUndo();
       if (ustrcmp(get_extension(path),"hbm") == 0)
       { 
        PACKFILE *HBMfile = pack_fopen(path,"rp"); //read packed hbm
        pack_fread(&BorderCol,sizeof(unsigned char),HBMfile);
        for(i=0;i<FieldSizeX;i++) pack_fread(&BmpData[i][0],FieldSizeY,HBMfile);
        pack_fclose(HBMfile);
       }
       else if (ustrcmp(get_extension(path),"prg") == 0)
       {
        LoadPrg(path);
       }
       else if (ustrcmp(get_extension(path),"bmp") == 0)
       {
        LoadBMP();
       }
      }
     }

void ConvBackBMP(BITMAP *bitmap) //convert a pointed (referenced) bmp into C64 colour Bitmap-data
     {
      unsigned int detcol;
      for (int i = 0;i<FieldSizeX;i++)
      {
       for (int j = 0;j<FieldSizeY;j++)
       {
        detcol = GetMatchCol(getpixel(bitmap, i,j));   //detect colour , getr(),getg(),getb() for calculations
        BmpData[i][j] = detcol; //if (detcol == BGCol) {BmpData[i][j] = 16;} else {BmpData[i][j] = detcol;}
       } 
      }
     }

void LoadBMP()  //loading and converting (cut/stertch/colour) the BMP file to BmpData 
     {
      BITMAP *ImportBMP; 
      ImportBMP = load_bmp(path,BMPalette);
      ConvBackBMP(ImportBMP);
      destroy_bitmap(ImportBMP);
      BorderCol=DefaultBorderCol; BorderDrawer();
     }



unsigned int GetMatchCol(unsigned int inputcol)  //get the code of the closest c64 colour for the given RGB colour
        {
         int red = getr(inputcol), green = getg(inputcol), blue = getb(inputcol), rdelta, gdelta, bdelta, resultcol, mintemp=256;
         int sumerrors[16]; //comparable errors for C64 palette colours to decide which fits the best
         for (int i=0;i<16;i++) //comparing through the palette for error-deltas, getting the C64 colour with smallest error
         {
          rdelta = abs(getr(C64Palette[i])-red);
          gdelta = abs(getg(C64Palette[i])-green);
          bdelta = abs(getb(C64Palette[i])-blue);
          sumerrors[i]=rdelta+gdelta+bdelta;
          if (sumerrors[i] < mintemp) {mintemp=sumerrors[i];resultcol=i;}
         }        
         return resultcol;
        }

 //------------------------------------saver (.hbm)----------------

void SaveHBM(char *filepath)
      {
       //write a new packed file with the graphic-data  
        PACKFILE *HBMfile = pack_fopen(filepath,"wp");
        pack_fwrite(&BorderCol,sizeof(unsigned char),HBMfile);
        for(int i=0;i<FieldSizeX;i++) pack_fwrite(&BmpData[i][0],FieldSizeY,HBMfile);
        pack_fclose(HBMfile);
      }

int Save()
    {
     remove_timer();
trial:
     replace_extension(path,path,"hbm",sizeof(path));
     if (file_select_ex("Give a name and path for the .hbm file",path,"hbm",480,400,400) != 0)
     {
      fix_filename_slashes(path);
      replace_extension(path,path,"hbm",sizeof(path));
      if (exists(path))
      {
       if (alert("File exists. Overwrite?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1) SaveHBM(path);
       else goto trial;
      }
      else SaveHBM(path);
     }
     SetTimer();
     return D_O_K;
    }

//-------------------------------------------.prg (.bmp/png?) converter & saver -------------
int ConvertChr(int Charnum) //convert the selected Char's layers to C64 format in ConvBmp[8][64]
     {
      int k, ipos=SelChrPosX(Charnum), jpos=SelChrPosY(Charnum), colindex = 0, coltest, mempos, bytepos, Chrx, Chry;
      for (int colour=0; colour <= 15; colour++)
      {
       coltest=0; //counts, how mutch times the actually tested colour appeared
       for (int i=ipos; i<(ipos+ChrSizeX); i++)
       {
        for (int j=jpos; j<(jpos+ChrSizeY); j++)
        {
         Chrx= i-ipos; Chry= j-jpos;
         mempos = Chry; bytepos = Chrx-(Chrx/8)*8;
         if ( BmpData[i][j]==colour) // || (BmpData[i][j]==16 && colour==BGCol) )
         { //set the Char-pixel for the actual colour
          /*if (colour==16)
          {
           for (k=0;k<=colindex+1;k++) if(ConvChr[k][8]==BGCol)break;
           if (k<=colindex)
           { //background colour is used already
            ConvChr[k][mempos] |= pow2[bytepos];
           }
           else
           { //background colour is new
            coltest++; //counts, how mutch times the currently tested colour appeared
            ConvChr[colindex][mempos] |= pow2[bytepos];
           }
          }
          else
          {*/
           coltest++; //counts, how mutch times the currently tested colour appeared
           ConvChr[colindex][mempos] |= pow2[bytepos];
          //}
         }
         else
         { //reset the Char-pixel for current colour
          ConvChr[colindex][mempos] &= 255-pow2[bytepos];
         }
        }
       }
       if (coltest > 0) {                      //if the colour occured, goes to next Char-layer
                         ConvChr[colindex][8]= colour; //and store the current colour to Char layer's byte 64
                         colindex++;                                   //and even stored the index of the overlay Char
                        }                                              //in the upper nybble of the colourbyte
      }
      ClashCnt[Charnum]=colindex; //register clash/colourcount
      ColorMap[Charnum]=ConvChr[0][8]*16+ConvChr[1][8];
      return colindex; //returns number of used layers (colours)
     }

void DetectClashes()
     {
      int colournum, i, GridClashX, GridClashY; //signs how many colours (layers) the specific Char contains      
      for (i=0 ; i<NoOfChrX*NoOfChrY ; i++) ConvertChr(i);
     }

void SavePRG(char *filepath)   //write a new binary file with the graphic-data  
     {
      int colournum, i, GridClashX, GridClashY; //signs how many colours (layers) the specific Char contains      
      PACKFILE *PRGFile = pack_fopen(filepath,"w");
      pack_fwrite(&PrgStartAdd,2,PRGFile);
      for (i=0 ; i<NoOfChrX*NoOfChrY ; i++)
      {
       ConvertChr(i);
       pack_fwrite(&ConvChr,8,PRGFile);
      }
      pack_fwrite(&ColorMap,1000,PRGFile);
      GridDrawer(GridCol);
      pack_fwrite(&BorderCol,sizeof(unsigned char),PRGFile); //the last byte of the prg is the background colour
      pack_fclose(PRGFile);
     } 

void SaveEXE(char *filepath)   //write a new binary file with the graphic-data  
     {
      int colournum, i,j, GridClashX, GridClashY, titlend=0; //signs how many colours (layers) the specific Char contains      
      PACKFILE *PRGFile = pack_fopen(filepath,"w");
      char * filename=get_filename(filepath);
      //textprintf_ex(screen,font,TxtInfoPosX, TxtInfoPosY-14,0x80FFFF,0x0000FF,"%s",filename);
      //pack_fwrite(&BASICstartAdd,2,PRGFile);  //already contained in the array!
      pack_fwrite(&bmpExe,0x201-40,PRGFile);

      i=0; if (strlen(filename)<24+4) for(;i<(24+4-strlen(filename))/2;i++) pack_putc(0x20,PRGFile);
      for (j=0 ;i<40;i++,j++) //pack_fwrite(&filename,5,PRGFile);
      {
       if (filename[j]==0 or filename[j]=='.') titlend=1;
       if (titlend==0) pack_putc(toupper(filename[j]),PRGFile);
       else  pack_putc(0,PRGFile);
      }
      for (i=0 ; i<NoOfChrX*NoOfChrY ; i++)
      {
       ConvertChr(i);
       pack_fwrite(&ConvChr,8,PRGFile);
      }
      pack_fwrite(&ColorMap,1000,PRGFile);
      GridDrawer(GridCol);
      pack_fwrite(&BorderCol,sizeof(unsigned char),PRGFile); //the last byte of the prg is the background colour
      pack_fclose(PRGFile);
     } 

int ExportC64() //save the data in C64 Bitmap (Art Studio 1) format .prg binary file (load-address included)
    {
     remove_timer();
trial2:
     replace_extension(path,path,"prg",sizeof(path));
     if (file_select_ex("Give a name and path for the .prg file",path,"prg",480,400,400) != 0)
     {
      fix_filename_slashes(path);
      replace_extension(path,path,"prg",sizeof(path));
      if (exists(path))
      {
       if (alert("File exists. Overwrite?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1) SavePRG(path);
       else goto trial2;
      }
      else SavePRG(path);
     }
     SetTimer();
     return D_O_K;
    }

int ExportExe () //convert to C64 executalbe
    {
     remove_timer();
trial2:
     replace_extension(path,path,"prg",sizeof(path));
     if (file_select_ex("Give a name and path for the .prg file",path,"prg",480,400,400) != 0)
     {
      fix_filename_slashes(path);
      replace_extension(path,path,"prg",sizeof(path));
      if (exists(path))
      {
       if (alert("File exists. Overwrite?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1) SaveEXE(path);
       else goto trial2;
      }
      else SaveEXE(path);
     }
     SetTimer();
     return D_O_K;
    }

//---------------------------------------------------------------------------------
void SaveBMPfile(char *filepath)
     {
      save_bmp(filepath,Preview,BMPalette);
     }

int SaveBMP()
    {
     PrevRightClick = 0;
     remove_timer();
trial3:
     replace_extension(path,path,"bmp",sizeof(path));
     if (file_select_ex("You can now save a bitmap, give name and path.",path,"bmp",480,400,400) != 0)
     {
      fix_filename_slashes(path);
      replace_extension(path,path,"bmp",sizeof(path));  // leaving this lets the user rename & set the picture-type to bmp/pcx/tga
      if (exists(path))
      {
       if (alert("File exists. Overwrite?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1) SaveBMPfile(path);
       else goto trial3;
      }
      else SaveBMPfile(path);
     }
     SetTimer();
     return D_O_K;
    }

//------------------------
int ToggleGrid()
    {
     remove_timer();
     if (EnableGrid==false) { EnableGrid=true; GridDrawer(GridCol); }
     else { EnableGrid=false; GridDrawer(0); }
     TextDispl();
     SetTimer();
     return D_O_K;
    }

int ClashTest()
    {
     remove_timer();
     if (ClashTestMode==false) { ClashTestMode=true; DetectClashes(); }
     else { ClashTestMode=false; GridDrawer(GridCol); }
     TextDispl();
     SetTimer();
     return D_O_K;
    }

/*int ToggleWide()
    {
     remove_timer();
     if (WideMode==false) 
     { 
      WideMode=true;     
      set_gfx_mode(GFX_AUTODETECT_WINDOWED, 1280, 1000, 0, 0);
     }
     else 
     { 
      set_gfx_mode(GFX_AUTODETECT_WINDOWED, GuiSizeX, GuiSizeY, 0, 0);
      WideMode=false; 
     }
     GridDrawer(GridCol); 
     TextDispl();
     SetTimer();
     return D_O_K;
    }*/

int ToggleBriGrid()
    {
     remove_timer();
     if (EnableBriGrid==false) { EnableBriGrid=true; GridDrawer(GridCol); }
     else { EnableBriGrid=false; GridDrawer(GridCol); }
     TextDispl();
     SetTimer();
     return D_O_K;
    }

//==================================================Pencil Tools==============================================
int F1er()
    {
     PenDiameter = 4; DispToolBar();
     return D_O_K;
    }

int F2er()
    {
     PenDiameter = 8; DispToolBar();
     return D_O_K;
    }

int F3er()
    {
     PenDiameter = 16; DispToolBar();
     return D_O_K;
    }

int F4er()
    {
     PenDiameter = 32; DispToolBar();
     return D_O_K;
    }

int F5er()
    {
     DrawMode = 1;DispToolBar();
     return D_O_K;
    }

int F6er()
    {
     if (DrawMode == 2) (RectFill == 1)? RectFill=0:RectFill=1; else DrawMode=2;DispToolBar();
     return D_O_K;
    }

int F7er()
    {
     if (DrawMode == 3) (CircFill == 1)? CircFill=0:CircFill=1; else DrawMode=3;DispToolBar();
     return D_O_K;
    }

int F8er()
    {
     DrawMode = 4; DispToolBar();
     return D_O_K;
    }

int F9er()
    {
     (DitherMode < 2)? DitherMode++:DitherMode=0;
     DispToolBar();
     return D_O_K;
    }

int F10er()
    {
     (Dither2Col!=1)?Dither2Col=1:Dither2Col=0; DispToolBar();
     return D_O_K;
    }

int F11er()
    {
     DrawMode=5; DispToolBar();
     return D_O_K;
    }

//===============================================UNDO/REDO=================================
void CopyBmpToUndo()
     {
      for (int i=0;i<FieldSizeX;i++)
      {
       for (int j=0;j<FieldSizeY;j++)
       {
        UndoStore[UndoPointer][i][j]= BmpData[i][j];
       }
      }
      //UndoStore[UndoPointer][0][FieldSizeY]= BGCol;
     }

void StoreUndo()
     {
      CopyBmpToUndo();
      (UndoPointer < 15) ? UndoPointer++: UndoPointer=0;
      RedoMax=UndoPointer;
      if (UndoPointer == UndoBase)
      {
       UndoBase++; UndoBase &= 0x0F;
       RedoMax=(UndoBase-1) & 0x0F;
      }
     }

int Zeer()
    {
     if (key_shifts & KB_CTRL_FLAG)
     {
      if (key_shifts & KB_SHIFT_FLAG)
      {
       remove_timer();
       Redoer();
       SetTimer();
      }
      else
      {
       remove_timer();
       Undoer();
       SetTimer();
      }
     }
     return D_O_K; //for keyboard-launcher
    }

int Yeer()
    {
     if (key_shifts & KB_CTRL_FLAG)
     {
      remove_timer();
      Redoer();
      SetTimer();
     }
     return D_O_K; //for keyboard-launcher
    }

void GetBmpFromUndo()
     {
      for (int i=0;i<FieldSizeX;i++)
      {
       for (int j=0;j<FieldSizeY;j++)
       {
        BmpData[i][j] = UndoStore[UndoPointer][i][j];
       }
      }
      //BGCol=UndoStore[UndoPointer][0][FieldSizeY];
      GridDrawer(GridCol);PrevDrawer();
     }

void Undoer()
     {
      CopyBmpToUndo();
      if(UndoPointer != UndoBase)
      {
       UndoPointer--; UndoPointer &= 0x0F;
       GetBmpFromUndo();
      }
     }

void Redoer()
     {
      if(UndoPointer != RedoMax)
      {
       UndoPointer++; UndoPointer &= 0x0F;
       GetBmpFromUndo();
      }
     }

//===========================================EOF=================================================================================
